@@ -9,13 +9,13 @@
 #include <stdlib.h>
 #include <termios.h>
 #include <unistd.h>
-#include <wctype.h>
 
-#include <algorithm>
 #include <cwchar>
+#include <deque>
+#include <functional>
 #include <memory>
 #include <string>
-#include <type_traits>
+#include <utility>
 #include <vector>
 
 #include "ast.h"
@@ -32,6 +32,7 @@
 #include "io.h"
 #include "job_group.h"
 #include "maybe.h"
+#include "operation_context.h"
 #include "parse_constants.h"
 #include "parse_util.h"
 #include "parser.h"
@@ -41,7 +42,6 @@
 #include "timer.h"
 #include "tokenizer.h"
 #include "trace.h"
-#include "util.h"
 #include "wildcard.h"
 #include "wutil.h"
 
@@ -393,13 +393,9 @@ end_execution_reason_t parse_execution_context_t::run_function_statement(
     null_output_stream_t outs;
     string_output_stream_t errs;
     io_streams_t streams(outs, errs);
-    int err_code = 0;
-    maybe_t<int> err = builtin_function(*parser, streams, arguments, pstree, statement);
-    if (err) {
-        err_code = err.value();
-        parser->libdata().status_count++;
-        parser->set_last_statuses(statuses_t::just(err_code));
-    }
+    int err_code = builtin_function(*parser, streams, arguments, pstree, statement);
+    parser->libdata().status_count++;
+    parser->set_last_statuses(statuses_t::just(err_code));
 
     const wcstring &errtext = errs.contents();
     if (!errtext.empty()) {
@@ -454,7 +450,7 @@ end_execution_reason_t parse_execution_context_t::run_for_statement(
     auto var = parser->vars().get(for_var_name, ENV_DEFAULT);
     if (env_var_t::flags_for(for_var_name.c_str()) & env_var_t::flag_read_only) {
         return report_error(STATUS_INVALID_ARGS, header.var_name,
-                            L"for: %ls: cannot overwrite read-only variable", for_var_name.c_str());
+                            _(L"%ls: %ls: cannot overwrite read-only variable"), L"for", for_var_name.c_str());
     }
 
     auto &vars = parser->vars();
@@ -735,9 +731,23 @@ end_execution_reason_t parse_execution_context_t::handle_command_not_found(
         // but this mainly applies to EACCES. We could also feasibly get:
         // ELOOP
         // ENAMETOOLONG
+        if (err_code == ENOTDIR) {
+            // If the original command did not include a "/", assume we found it via $PATH.
+            auto src = get_source(statement.command);
+            if (src.find(L"/") == wcstring::npos) {
+                return this->report_error(
+                                          STATUS_NOT_EXECUTABLE, statement.command,
+                                          _(L"Unknown command. A component of '%ls' is not a directory. Check your $PATH."), cmd);
+            } else {
+                return this->report_error(
+                                          STATUS_NOT_EXECUTABLE, statement.command,
+                                          _(L"Unknown command. A component of '%ls' is not a directory."), cmd);
+            }
+        }
+
         return this->report_error(
-            STATUS_NOT_EXECUTABLE, statement,
-            _(L"Unknown command. '%ls' exists but is not an executable file."), cmd);
+                                  STATUS_NOT_EXECUTABLE, statement.command,
+                                  _(L"Unknown command. '%ls' exists but is not an executable file."), cmd);
     }
 
     // Handle unrecognized commands with standard command not found handler that can make better
@@ -766,7 +776,7 @@ end_execution_reason_t parse_execution_context_t::handle_command_not_found(
         buffer = L"fish_command_not_found";
         for (const wcstring &arg : event_args) {
             buffer.push_back(L' ');
-            buffer.append(escape_string(arg, ESCAPE_ALL));
+            buffer.append(escape_string(arg));
         }
         auto prev_statuses = parser->get_last_statuses();
 
@@ -781,7 +791,7 @@ end_execution_reason_t parse_execution_context_t::handle_command_not_found(
         error = _(L"Unknown command:");
         if (!event_args.empty()) {
             error.push_back(L' ');
-            error.append(escape_string(event_args[0], ESCAPE_ALL));
+            error.append(escape_string(event_args[0]));
         }
     }
 
@@ -791,7 +801,7 @@ end_execution_reason_t parse_execution_context_t::handle_command_not_found(
 
     // Here we want to report an error (so it shows a backtrace).
     // If the handler printed text, that's already shown, so error will be empty.
-    return this->report_error(STATUS_CMD_UNKNOWN, statement, error.c_str());
+    return this->report_error(STATUS_CMD_UNKNOWN, statement.command, error.c_str());
 }
 
 end_execution_reason_t parse_execution_context_t::expand_command(
@@ -825,7 +835,7 @@ end_execution_reason_t parse_execution_context_t::expand_command(
     // Complain if the resulting expansion was empty, or expanded to an empty string.
     // For no-exec it's okay, as we can't really perform the expansion.
     if (out_cmd->empty() && !no_exec()) {
-        return this->report_error(STATUS_ILLEGAL_CMD, statement,
+        return this->report_error(STATUS_ILLEGAL_CMD, statement.command,
                                   _(L"The expanded command was empty."));
     }
     return end_execution_reason_t::ok;
@@ -1011,7 +1021,8 @@ end_execution_reason_t parse_execution_context_t::determine_redirections(
         redirection_spec_t spec{oper->fd, oper->mode, std::move(target)};
 
         // Validate this spec.
-        if (spec.mode == redirection_mode_t::fd && !spec.is_close() && !spec.get_target_as_fd()) {
+        if (spec.mode == redirection_mode_t::fd && !spec.is_close() &&
+                !spec.get_target_as_fd().has_value()) {
             const wchar_t *fmt =
                 _(L"Requested redirection to '%ls', which is not a valid file descriptor");
             return report_error(STATUS_INVALID_ARGS, redir_node, fmt, spec.target.c_str());
@@ -1085,7 +1096,7 @@ end_execution_reason_t parse_execution_context_t::apply_variable_assignments(
     for (const ast::variable_assignment_t &variable_assignment : variable_assignment_list) {
         const wcstring &source = get_source(variable_assignment);
         auto equals_pos = variable_assignment_equals_pos(source);
-        assert(equals_pos);
+        assert(equals_pos.has_value());
         const wcstring variable_name = source.substr(0, *equals_pos);
         const wcstring expression = source.substr(*equals_pos + 1);
         completion_list_t expression_expanded;
@@ -1520,8 +1531,11 @@ end_execution_reason_t parse_execution_context_t::eval_node(const ast::job_list_
                                   INFINITE_FUNC_RECURSION_ERR_MSG, func_name.c_str());
     }
 
-    // Check for stack overflow. The TOP check ensures we only do this for function calls.
-    if (associated_block->type() == block_type_t::top && parser->function_stack_is_overflowing()) {
+    // Check for stack overflow in case of function calls (regular stack overflow) or string
+    // substitution blocks, which can be recursively called with eval (issue #9302).
+    if ((associated_block->type() == block_type_t::top &&
+         parser->function_stack_is_overflowing()) ||
+        (associated_block->type() == block_type_t::subst && parser->is_eval_depth_exceeded())) {
         return this->report_error(STATUS_CMD_ERROR, job_list, CALL_STACK_LIMIT_EXCEEDED_ERR_MSG);
     }
     return this->run_job_list(job_list, associated_block);
