@@ -4,21 +4,20 @@
 // keeping an internal representation of the current screen contents and trying to find the most
 // efficient way for transforming that to the desired screen content.
 //
-// IWYU pragma: no_include <cstddef>
 #include "config.h"
 
 #include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <termios.h>
-#include <time.h>
 #include <unistd.h>
 
+#include <cstdint>
 #include <cstring>
 #include <cwchar>
 
 #if HAVE_CURSES_H
-#include <curses.h>
+#include <curses.h>  // IWYU pragma: keep
 #elif HAVE_NCURSES_H
 #include <ncurses.h>
 #elif HAVE_NCURSES_CURSES_H
@@ -38,10 +37,12 @@
 #include "env.h"
 #include "fallback.h"  // IWYU pragma: keep
 #include "flog.h"
+#include "global_safety.h"
 #include "highlight.h"
 #include "output.h"
 #include "pager.h"
 #include "screen.h"
+#include "termsize.h"
 
 /// The number of characters to indent new blocks.
 #define INDENT_STEP 4u
@@ -74,6 +75,12 @@ static size_t try_sequence(const char *seq, const wchar_t *str) {
     return 0;  // this should never be executed
 }
 
+static bool midnight_commander_hack = false;
+
+void screen_set_midnight_commander_hack() {
+    midnight_commander_hack = true;
+}
+
 /// Returns the number of columns left until the next tab stop, given the current cursor position.
 static size_t next_tab_stop(size_t current_line_width) {
     // Assume tab stops every 8 characters if undefined.
@@ -259,6 +266,13 @@ maybe_t<size_t> escape_code_length(const wchar_t *code) {
     return found ? maybe_t<size_t>{esc_seq_len} : none();
 }
 
+wcstring screen_clear() {
+    if (clear_screen) {
+        return str2wcstring(clear_screen);
+    }
+    return wcstring{};
+}
+
 size_t layout_cache_t::escape_code_length(const wchar_t *code) {
     assert(code != nullptr);
     if (*code != L'\x1B') return 0;
@@ -267,7 +281,7 @@ size_t layout_cache_t::escape_code_length(const wchar_t *code) {
     if (esc_seq_len) return esc_seq_len;
 
     auto found = ::escape_code_length(code);
-    if (found) {
+    if (found.has_value()) {
         this->add_escape_code(wcstring(code, *found));
         esc_seq_len = *found;
     }
@@ -871,8 +885,12 @@ void screen_t::update(const wcstring &left_prompt, const wcstring &right_prompt,
                 clear_remainder = prev_width > current_width;
             }
         }
+
+        // We unset the color even if we don't clear the line.
+        // This means that we switch background correctly on the next,
+        // including our weird implicit bolding.
+        set_color(highlight_spec_t{});
         if (clear_remainder && clr_eol) {
-            set_color(highlight_spec_t{});
             this->move(current_width, static_cast<int>(i));
             this->write_mbs(clr_eol);
         }
@@ -904,7 +922,11 @@ void screen_t::update(const wcstring &left_prompt, const wcstring &right_prompt,
 
     // Also move the cursor to the beginning of the line here,
     // in case we're wrong about the width anywhere.
-    this->move(0, 0);
+    // Don't do it when running in midnight_commander because of
+    // https://midnight-commander.org/ticket/4258.
+    if (!midnight_commander_hack) {
+        this->move(0, 0);
+    }
 
     // Clear remaining lines (if any) if we haven't cleared the screen.
     if (!has_cleared_screen && need_clear_screen && clr_eol) {
@@ -1258,7 +1280,7 @@ void screen_t::reset_abandoning_line(int screen_width) {
             std::string dim = fish_tparm(const_cast<char *>(enter_dim_mode));
             if (!dim.empty()) {
                 // Use dim if they have it, so the color will be based on their actual normal
-                // color and the background of the termianl.
+                // color and the background of the terminal.
                 abandon_line_string.append(str2wcstring(dim));
                 justgrey = false;
             }
